from typing import TypeVar, Generic, List, Optional
from edge import Edge

V = TypeVar('V')  # тип вершин графа


class Graph(Generic[V]):
    def __init__(self, vertices: List[V] = []) -> None:
        self._vertices: List[V] = vertices
        self._edges: List[List[Edge]] = [[] for _ in vertices]

    @property
    def vertex_count(self) -> int:
        return len(self._vertices)  # Количество вершин

    @property
    def edge_count(self) -> int:
        return sum(map(len, self._edges))  # Количество ребер

    # Добавляем вершину в граф и возвращаем ее индекс
    def add_vertex(self, vertex: V) -> int:
        self._vertices.append(vertex)
        self._edges.append([])  # Добавляем пустой список для ребер
        return self.vertex_count - 1  # Возвращаем индекс по добавленным вершинам

    # Это ненаправленный граф,
    # поэтому мы всегда добавляем вершины в обоих направлениях
    def add_edge(self, edge: Edge) -> None:
        self._edges[edge.u].append(edge)
        self._edges[edge.v].append(edge.reversed())

    # Добавляем ребро, используя индексы вершин (удобный метод)
    def add_edge_by_indices(self, u: int, v: int) -> None:
        edge: Edge = Edge(u, v)
        self.add_edge(edge)

    # Добавляем ребро, просматривая индексы вершин (удобный метод)
    def add_edge_by_vertices(self, first: V, second: V) -> None:
        u: int = self._vertices.index(first)
        v: int = self._vertices.index(second)
        self.add_edge_by_indices(u, v)

    # Поиск вершины по индексу
    def vertex_at(self, index: int) -> V:
        return self._vertices[index]

    # Поиск индекса вершины в графе
    def index_of(self, vertex: V) -> int:
        return self._vertices.index(vertex)

    # Поиск вершин, с которыми связана вершина с заданным индексом
    def neighbors_for_index(self, index: int) -> List[V]:
        return list(map(self.vertex_at, [e.v for e in self._edges[index]]))

    # Поиск индекса вершины; возвращает ее соседей (удобный метод)
    def neighbors_for_vertex(self, vertex: V) -> List[V]:
        return self.neighbors_for_index(self.index_of(vertex))

    # Возвращает все ребра, связанные с вершиной, имеющей заданный индекс
    def edges_for_index(self, index: int) -> List[Edge]:
        return self._edges[index]

    # Поиск индекса вершины; возвращает ее ребра (удобный метод)
    def edges_for_vertex(self, vertex: V) -> List[Edge]:
        return self.edges_for_index(self.index_of(vertex))

    # Упрощенный красивый вывод графа
    def __str__(self) -> str:
        desc: str = ""
        for i in range(self.vertex_count):
            desc += f"{self.vertex_at(i)} -> {self.neighbors_for_index(i)}\n"
        return desc


if __name__ == "__main__":
    # тест простейшей графовой конструкции
    city_graph: Graph[str] = Graph(["Сиэтл", "Сан-Франциско", "Лос-Анджелес", "Риверсайд",
                                    "Феникс", "Чикаго", "Бостон", "Нью-Йорк",
                                    "Атланта", "Майами", "Даллас", "Хьюстон", "Детройт",
                                    "Филадельфия", "Вашингтон"])
    city_graph.add_edge_by_vertices("Сиэтл", "Чикаго")
    city_graph.add_edge_by_vertices("Сиэтл", "Сан-Франциско")
    city_graph.add_edge_by_vertices("Сан-Франциско", "Риверсайд")
    city_graph.add_edge_by_vertices("Сан-Франциско", "Лос-Анджелес")
    city_graph.add_edge_by_vertices("Лос-Анджелес", "Риверсайд")
    city_graph.add_edge_by_vertices("Лос-Анджелес", "Феникс")
    city_graph.add_edge_by_vertices("Риверсайд", "Феникс")
    city_graph.add_edge_by_vertices("Риверсайд", "Чикаго")
    city_graph.add_edge_by_vertices("Феникс", "Даллас")
    city_graph.add_edge_by_vertices("Феникс", "Хьюстон")
    city_graph.add_edge_by_vertices("Даллас", "Чикаго")
    city_graph.add_edge_by_vertices("Даллас", "Атланта")
    city_graph.add_edge_by_vertices("Даллас", "Хьюстон")
    city_graph.add_edge_by_vertices("Хьюстон", "Атланта")
    city_graph.add_edge_by_vertices("Хьюстон", "Майами")
    city_graph.add_edge_by_vertices("Атланта", "Чикаго")
    city_graph.add_edge_by_vertices("Атланта", "Вашингтон")
    city_graph.add_edge_by_vertices("Атланта", "Майами")
    city_graph.add_edge_by_vertices("Майами", "Вашингтон")
    city_graph.add_edge_by_vertices("Чикаго", "Детройт")
    city_graph.add_edge_by_vertices("Детройт", "Бостон")
    city_graph.add_edge_by_vertices("Детройт", "Вашингтон")
    city_graph.add_edge_by_vertices("Детройт", "Нью-Йорк")
    city_graph.add_edge_by_vertices("Бостон", "Нью-Йорк")
    city_graph.add_edge_by_vertices("Нью-Йорк", "Филадельфия")
    city_graph.add_edge_by_vertices("Филадельфия", "Вашингтон")
    print(city_graph)
